import random
import string
import time
from typing import List, Any, Optional

LINE_UP = '\033[1A'
LINE_CLEAR = '\x1b[2K'



def short_scin(
        fl: float,
        precision:int=  1,
) -> str:
    """ short (compressed) scientific notation for floats """
    sh = f'{fl:.{precision}E}'
    sh = sh.replace('+0','')
    sh = sh.replace('+','')
    sh = sh.replace('-0','-')
    sh = sh.replace('E','e')
    return sh


def float_to_str(
        num: float,
        width: int=     7,
        fill: bool=     True,
) -> str:
    """ returns nice string from float, always of given width """

    if width < 5: width = 5
    scientific_decimals = width-6 if width>6 else 0

    fstr = f'{num:.{scientific_decimals}E}'
    if 1000 > num > 0.0001 or num == 0.0:
        fstr = str(num)[:width]

    if fill and len(fstr)<width:
        fstr += ' '*(width-len(fstr))

    return fstr


def stamp(
        year: bool=             False,
        month: bool=            True,
        day: bool=              True,
        hour: bool=             True,
        minutes: bool=          True,
        letters: Optional[int]= 3,
) -> str:
    """ prepares timestamp string """

    random.seed(time.time())

    time_format = ''
    if year:            time_format += '%y'
    if month:           time_format += '%m'
    if day:             time_format += '%d'
    if time_format:     time_format += '_'
    if hour:            time_format += '%H'
    if minutes:         time_format += '%M'

    stp = ''
    if time_format:
        stp = time.strftime(time_format)

    if letters:
        if stp: stp += '_'
        stp += ''.join([random.choice(string.ascii_letters) for _ in range(letters)])

    return stp


def list_str(ls: List[Any], limit:Optional[int]=200) -> str:
    """ nice string of given list """
    lstr = [str(e) for e in ls]
    lstr = '; '.join(lstr)
    if limit: lstr = lstr[:limit]
    return lstr


def print_nested_dict(
        d: dict,
        ind_scale: int=     2,
        line_limit: int=    100,
) -> None:
    """ prints nice string of nested dict """

    types = {
        dict:   'D',
        list:   'L',
        tuple:  'T',
        str:    'S'}

    def __prn_root(root:dict, ind:int):

        spacer = ' ' * ind * ind_scale

        for key in sorted(list(root.keys())):

            value = root[key]
            tp = types.get(type(value),'')
            value_len_str = str(len(value)) if tp in types.values() else ''

            line = ''
            if tp != 'D':
                line = repr(value)
                if line_limit and len(line)>line_limit:
                    line = f'{line[:line_limit]} ..'
                line = f' : {line}'

            type_len_nfo = f' [{tp}.{value_len_str}]' if tp else ''
            print(f'{spacer}{key}{type_len_nfo}{line}')

            if type(value) is dict:
                __prn_root(root=root[key], ind=ind+1)

    __prn_root(root=d, ind=0)


def printover(sth, clear:int=10) -> None:
    """ prints line over line """
    cls = '' + ' ' * clear
    print(f'\r{sth}{cls}', end='', flush=True)


def printover_terminal(sth) -> None:
    """ prints line over line (terminal alternative) """
    print(sth)
    print(LINE_UP, end=LINE_CLEAR)


def progress_ (
        current,                # current progress
        total,                  # total
        prefix: str=    '',     # prefix string
        suffix: str=    '',     # suffix string
        length: int=    20,
        fill: str=      '█',
) -> None:
    """ terminal progress bar """
    prog = current / total
    if prog > 1: prog = 1
    filled_length = int(length * prog)
    bar = fill * filled_length + '-' * (length - filled_length)
    printover(f'{prefix} |{bar}| {prog*100:.1f}% {suffix}')
    if prog == 1: print()


from typing import Optional

from pypaq.lipytools.pylogger import get_pylogger
from pypaq.lipytools.files import prep_folder, r_json, w_json
from pypaq.pms.base import POINT


# Configuration Manager, keeps configuration in attributes of self, manages config file (json)
class ConfigManager:

    def __init__(
            self,
            file_FP: str,                           # full path to config file
            config_init: Optional[POINT]=   None,
            override_with_saved: bool=      True,   # file (if exists) will override given initial config
            logger=                         None,
            loglevel=                       20):

        if not logger:
            logger = get_pylogger(name='ConfigManager', level=loglevel)
        self._logger = logger

        self._file = file_FP
        prep_folder(self._file)
        self._logger.info(f'*** ConfigManager *** inits, config file: {self._file}')

        _conf = {}

        if config_init is not None:
            self._logger.info(f'> got initial config: {config_init}')
            _conf.update(config_init)

        file_config = r_json(self._file)
        if file_config:
            self._logger.info(f'> got file config:    {file_config}')
            if override_with_saved:
                _conf.update(file_config)

        self.__config = _conf
        w_json(_conf, file_FP)

    # sets attribute
    def __setattr__(self, key, value):
        if key.startswith('_'):
            super().__setattr__(key, value)
        else:

            # catch once
            _conf = self.__config
            _file = self._file
            _logger = self._logger

            # first update with file (no changes monitoring, since usually *setting* changes sth)
            file_config = r_json(_file)
            _conf.update(file_config)

            new_attribute = key not in _conf
            new_value = not new_attribute and _conf[key] != value
            if new_attribute or new_value:
                _conf[key] = value
                if new_attribute:   _logger.info(f'set new attribute: {key}: {value}')
                if new_value:       _logger.info(f'set new value:     {key}: {value}')

            w_json(_conf, _file)

    # returns attribute value
    def __getattribute__(self, name):
        if name.startswith('_') or name in ['update_config','get_config']:
            return super().__getattribute__(name)
        else:

            # catch once
            _conf = self.__config
            _file = self._file

            # first update with file, with changes monitoring
            file_config = r_json(_file)
            changed_sth = False
            for k in file_config:
                if k not in _conf or k in _conf and _conf[k] != file_config[k]:
                    _conf[k] = file_config[k]
                    changed_sth = True

            if changed_sth:
                w_json(_conf, _file)

            return _conf[name]

    # update with give dictionary
    def update_config(self, dct:POINT) -> None:

        # catch once
        _conf = self.__config
        _file = self._file
        _logger = self._logger

        # first update with file (no changes monitoring, since usually *updating* changes sth)
        file_config = r_json(_file)
        _conf.update(file_config)

        if dct:
            _conf.update(dct)
            _logger.info(f'updated configuration with: {dct}')

        w_json(_conf, _file)

    def get_config(self) -> POINT:
        self.update_config({})
        c = {}
        c.update(self.__config)
        return c

    def __str__(self):
        return str(self.get_config())


import Levenshtein


# returns Levenshtein-distance of two strings
def lev_dist(source :str, target :str):
    return Levenshtein.distance(source, target)

# returns Levenshtein-distance of two lists (or strings)
def lev_distL(source :list or str, target :list or str):

    # same
    if len(source) == len(target):
        if not sum([1 for e in zip(source,target) if e[0] == e[1]]):
            return 0

    # prepare a matrix
    slen, tlen = len(source), len(target)
    dist = [[0 for _ in range(tlen+1)] for _ in range(slen+1)]
    for i in range(slen+1): dist[i][0] = i
    for j in range(tlen+1): dist[0][j] = j

    # count distance
    for i in range(slen):
        for j in range(tlen):
            cost = 0 if source[i] == target[j] else 1
            dist[i + 1][j + 1] = min(dist[i][j + 1] + 1,    # delete
                                     dist[i + 1][j] + 1,    # insert
                                     dist[i][j] + cost)     # substitute

    return dist[-1][-1]

# selects two furthest (with lev_dist) of three sentences
def two_most_distanced(sa :str,sb :str,sc :str):
    palev = [[lev_dist(*pair), pair] for pair in [[sa, sb], [sb, sc], [sc, sa]]]
    return sorted(palev)[-1][1]


from abc import ABC, abstractmethod
from multiprocessing import cpu_count, Process, Queue, Value
from queue import Empty
import psutil
import time
from typing import Any, Optional

from pypaq.exception import PyPaqException
from pypaq.lipytools.pylogger import get_pylogger


class MPythonException(PyPaqException):
    pass


# message sent between processes via Ques (my que)
class QMessage:
    def __init__(self, type:str, data:Optional[Any]=None):
        self.type = type
        self.data = data

    def __str__(self):
        return f'{self.__class__.__name__}, type:{self.type}, data:{self.data}'


# https://github.com/vterron/lemon/commit/9ca6b4b1212228dbd4f69b88aaf88b12952d7d6f
class SharedCounter:

    def __init__(self, n:int=0):
        self.count = Value('i', n)

    def increment(self, n:int=1):
        with self.count.get_lock():
            self.count.value += n

    @property
    def value(self) -> int:
        return self.count.value


# my que
class Que:

    def __init__(self):
        self.q = Queue()
        self.size = SharedCounter(0)

    def put(self, msg:QMessage, **kwargs):
        if not isinstance(msg, QMessage):
            raise MPythonException(f'\'msg\' should be type of QMessage, but is {type(msg)}')
        self.size.increment(1)
        self.q.put(msg, **kwargs)

    # does not raise Empty exception, but returns None in case no QMessage
    def get(self,
            block: bool=                True,
            timeout: Optional[float]=   None,
            ) -> Optional[QMessage]:
        try:
            msg = self.q.get(block=block, timeout=timeout)
            self.size.increment(-1)
            if not isinstance(msg, QMessage):
                raise MPythonException(f'\'msg\' should be type of QMessage, but is {type(msg)}')
            return msg
        except Empty:
            return None

    def empty(self) -> bool:
        return not self.qsize()

    def qsize(self) -> int:
        return self.size.value


# Exception Managed Subprocess - with basic exceptions management
class ExSubprocess(Process, ABC):

    def __init__(
            self,
            ique: Optional[Que]=        None,   # input que
            oque: Optional[Que]=        None,   # output que
            id: Optional[int or str]=   None,   # unique id to identify the subprocess, if not given takes from Process name
            raise_unk_exception=        True,   # raises exception other than KeyboardInterrupt
            logger=                     None,
            loglevel=                   30):

        ABC.__init__(self)
        Process.__init__(self, target=self.__run)
        if id is None: id = self.name

        self.id = id

        if not logger:
            logger = get_pylogger(
                name=       self.id,
                folder=     None,
                level=      loglevel)
        self.logger = logger

        self.raise_unk_exception = raise_unk_exception

        self.ique = ique
        self.oque = oque

        self.logger.info(f'*** ExSubprocess *** id: {self.id} initialized')

    # process target method, wraps subprocess_method() with try / except
    def __run(self):
        try:
            self.logger.debug(f'> ExSubprocess ({self.id}, pid:{self.pid}) - started subprocess_method()')
            self.subprocess_method()
            self.logger.debug(f'> ExSubprocess ({self.id}, pid:{self.pid}) - finished subprocess_method()')
        except KeyboardInterrupt:
            self.__exception_handle('KeyboardInterrupt')
        except Exception as e:
            self.__exception_handle(f'other: {e}')
            if self.raise_unk_exception: raise e

    # method run in a subprocess, to be implemented
    @abstractmethod
    def subprocess_method(self): pass

    # when exception occurs, message with exception data is put on the output que
    def __exception_handle(self, name:str):
        msg = QMessage(
            type=   f'ex_{name}, ExSubprocess id: {self.id}, pid: {self.pid}',
            data=   self.id) # returns ID here to allow process identification
        if self.oque is not None:
            self.oque.put(msg)
        self.logger.warning(f'> ExSubprocess ({self.id}) halted by exception: {name}')
        self.after_exception_handle_run()

    # this method may be implemented and will be run after exception occurred
    def after_exception_handle_run(self): pass

    def kill(self):
        if self.alive: super().kill()
        while self.alive: time.sleep(0.01)

    def terminate(self):
        if self.alive: super().terminate()
        while self.alive: time.sleep(0.01)

    @property
    def alive(self):
        if self.closed: return False
        return self.is_alive()

    @property
    def closed(self):
        return self._closed

    # some process info
    def get_info(self) -> str:
        pid = self.pid if not self.closed else None
        pid_nfo = pid if pid is not None else '<pid:closed>'
        exitcode = self.exitcode if not self.closed else '<exitcode:closed>'
        mem_nfo = int(psutil.Process(pid).memory_info().rss / 1024 ** 2) if not exitcode else '-'
        nfo = f'{str(self)}, pid: {pid_nfo}, mem:{mem_nfo}MB, parent pid: {self._parent_pid}, alive: {self.alive}, exitcode: {exitcode}'
        return nfo


def sys_res_nfo():
    vm = psutil.virtual_memory()
    gb = 1024 ** 3
    return {
        'cpu_count':    cpu_count(),
        'cpu_used_%':   psutil.cpu_percent(interval=5), # over last 5 sec
        'mem_total_GB': vm.total / gb,
        'mem_used_%':   vm.percent}


import time
from typing import Dict, List, Tuple, Optional


class TimeRep:
    """ time reporting helper """

    def __init__(self):
        self.start_now()

    def start_now(self):
        """ resets and starts """
        self.tr: Dict[str, Tuple[float, Optional[TimeRep]]] = {}
        self.stime = time.time()
        self.stime_start = self.stime

    def log(self, interval_name:str, interval_tr:Optional["TimeRep"]=None):
        ct = time.time()
        self.tr[interval_name] = (ct - self.stime, interval_tr)
        self.stime = ct

    def get_report(self) -> Dict[str,float]:
        return {k: self.tr[k][0] for k in self.tr}

    @staticmethod
    def _get_line(n,t):
        return f'{n:30}: {t:9.3f}s'

    def _get_report_lines(self) -> List[str]:
        rl = []
        for n,(t, itr) in self.tr.items():
            rl.append(self._get_line(n,t))
            if itr:
                rl += [f'---{l}' for l in itr._get_report_lines()]
        return rl

    def __str__(self):
        s = "\n".join(self._get_report_lines())
        return s + f"\n{self._get_line('___total time', time.time() - self.stime_start)}"


from pypaq.exception import PyPaqException


# double hinge function _/**
def double_hinge(
        s_val: float,
        e_val: float,
        sf: float,
        ef: float,
        counter: int,
        max_count: int,
) -> float:
    """
    returns:
        - s_val in range <0;A> steps
        - linear interpolation from s_val to e_val in range <A;B>
        - e_val in range <A;max_count>

    counter (step) belongs to <0;max_count>
    A = max_count * sf
    B = max_count * (1-ef)
    0 <= A <= B <= max_count
    """
    if sf + ef > 1:
        raise PyPaqException(f'sf + ef cannot be higher than 1')

    x = counter / max_count         # where we are in time (x)
    y = (x-sf)/(1-sf-ef)            # how high we are
    if y<0: y=0
    if y>1: y=1
    val = s_val + (e_val-s_val)*y   # final value
    return val


from multiprocessing import Process, Queue
from functools import partial


def proc(f):
    """ non-blocking process
    does not return anything """

    def new_f(*args, **kwargs):
        p = Process(target=partial(f, *args, **kwargs))
        p.start()

    new_f.__name__ = f'@proc:{f.__name__}'
    return new_f

def proc_wait(f):
    """ blocking process
    does not return anything """

    def new_f(*args, **kwargs):
        p = Process(target=partial(f, *args, **kwargs))
        p.start()
        p.join()

    new_f.__name__ = f'@proc_wait:{f.__name__}'
    return new_f


class MProc(Process):
    """ helper class
    process with que: puts target result on que """

    def __init__(self, que:Queue, f, args, kwargs):
        super().__init__(target=self.proc_m)
        self.que = que
        self.f = f
        self.ag = args
        self.kw = kwargs

    def proc_m(self):
        res = self.f(*self.ag, **self.kw)
        self.que.put(res)

def proc_return(f):
    """ blocking, returning process """

    def new_f(*args, **kwargs):
        que = Queue()
        p = MProc(que=que, f=f, args=args, kwargs=kwargs)
        p.start()
        ret = que.get()
        return ret
    new_f.__name__ = f'@proc_return:{f.__name__}'

    return new_f

def proc_que(que):
    """ non-blocking process
    returns (via given que) """

    def wrap(f):

        def new_f(*args, **kwargs):
            p = MProc(que=que, f=f, args=args, kwargs=kwargs)
            p.start()

        new_f.__name__ = f'@qproc:{f.__name__}'
        return new_f

    return wrap


from functools import wraps
from inspect import getfullargspec
import time

from pypaq.lipytools.printout import stamp
from pypaq.pms.base import get_params


# decorator printing execution time report
def timing(f):
    def new_f(*args, **kwargs):
        stime = time.time()
        ret = f(*args, **kwargs)
        taken_sec = time.time() - stime
        taken = f'{taken_sec/60:.1f}min' if taken_sec > 100 else f'{taken_sec:.1f}sec'
        print(f'(@timing:{stamp(letters=None)}, taken {taken}) {f.__name__} finished')
        return ret
    new_f.__name__ = f'{f.__name__}:@timing'
    return new_f

# prints debug info about parameters and given args/kwargs of function/method
def args(f):
    def new_f(*args, **kwargs):
        ins = getfullargspec(f)
        no_val = '--'

        arL = ins.args
        if arL[0] == 'self': arL = arL[1:]                  # simple, BUT not 100% accurate, ..but who names param 'self'?

        defL = list(ins.defaults) if ins.defaults else []   # list of default values
        defL = [no_val]*(len(arL)-len(defL)) + defL         # pad them with no_val
        arL = [list(e) for e in zip(arL,defL)]              # zip together

        # append args from the beginning
        if args:
            arvL = args[1:]
            for ix in range(len(arvL)):
                v = arvL[ix]
                arL[ix].append(v)

        kwL = [[k,kwargs[k]] for k in kwargs]               # kwargs in list

        # get from kwargs params of f, append their values and remove from kwargs
        kremIXL = []
        for ix in range(len(kwL)):
            for e in arL:
                if kwL[ix][0]==e[0]:
                    kremIXL.append(ix)
                    e.append(kwL[ix][1])
        for ix in reversed(kremIXL): kwL.pop(ix)

        # add no value to not overridden params
        for e in arL:
            if len(e)<3: e.append(no_val)

        # calc columns widths, trim if
        kw = [10,10,10]
        for e in arL:
            for i in range(3):
                lse = len(str(e[i]))
                if lse>kw[i]: kw[i] = lse
        for e in kwL:
            for i in range(2):
                lse = len(str(e[i]))
                ki = i if not i else 2
                if lse > kw[ki]: kw[ki] = lse
        while sum(kw) > 120:
            mx = max(kw)
            for ix in range(3):
                if kw[ix] == mx: kw[ix] -= 1
        mx = max(kw)

        def s(p):
            r = str(p)
            if len(r) > mx: r = r[:mx-2] + '..'
            return r

        print(f'\n@args report: *****************************************************************************************')
        print(f'function __name__: {f.__name__}')
        print(f' > {"param":{str(kw[0])}s}  {"default":{str(kw[1])}s}  {"given":{str(kw[2])}s}')
        for e in arL:
            print(f'   {s(e[0]):{str(kw[0])}s}  {s(e[1]):{str(kw[1])}s}  {s(e[2]):{str(kw[2])}s}')
        if kwL: print(f' > **kwargs (not used by {f.__name__}):')
        for e in kwL:
            print(f'   {s(e[0]):{str(kw[0])}s}  {"":{str(kw[1])}s}  {s(e[1]):{str(kw[2])}s}')
        print('@args report finished *********************************************************************************\n')

        return f(*args, **kwargs)
    new_f.__name__ = f'{f.__name__}:@args'
    return new_f

# decorator for __init__(), automatically assigns __init__ args and kwargs to the object (self)
def autoinit(func):

    @wraps(func)
    def wrapper(self, *args, **kwargs):

        params = get_params(func)
        names = params['without_defaults'][1:]  # ..trim self
        defaults = params['with_defaults']

        for name, arg in list(zip(names, args)) + list(kwargs.items()):
            setattr(self, name, arg)

        # case, when default is given with args
        left_args = args[len(names):]
        for la,key in zip(left_args,defaults.keys()):
            defaults[key] = la

        for name in defaults:
            val = defaults[name]
            if not hasattr(self, name):
                setattr(self, name, val)

        func(self, *args, **kwargs)

    return wrapper


import pandas as pd
import numpy as np
from typing import Dict, Tuple

from pypaq.pytypes import NPL


# min, avg, max ..of num list
def mam(vals:NPL) -> Tuple[float,float,float]:
    if len(vals): return min(vals), sum(vals) / len(vals), max(vals)
    else:         return 0.0, 0.0, 0.0

# mean, std, min, max (from given list of values or np.arr)
def msmx(vals:NPL) -> Dict:

    arr = np.asarray(vals) if type(vals) is list else vals
    ret_dict = {
        'mean': float(np.mean(arr)),
        'std':  float(np.std(arr)),
        'min':  float(np.min(arr)),
        'max':  float(np.max(arr))}
    ret_dict['string'] = 'mean %.5f, std %.5f, min %.5f, max %.5f'%(ret_dict['mean'],ret_dict['std'],ret_dict['min'],ret_dict['max'])
    return ret_dict

# deep stats (with pandas)
def stats_pd(
        val_list :NPL,
        n_percentiles=  10) -> str:
    s = f'{pd.Series(val_list).describe(percentiles=[0.1*n for n in range(1,n_percentiles)])}'
    return s[:s.rfind('\n')]


from copy import deepcopy
from typing import List, Optional, Union, Dict, Set, Tuple

from pypaq.textools.text_metrics import lev_dist
from pypaq.pms.base import POINT, PSDD, PMSException
from pypaq.pms.paspa import PaSpa


def dict_diff(da:Dict, db:Dict) -> str:
    """ returns nice str information about dict differences: db against da """

    nfo = ''

    missing_keys = []
    for k in da:
        if k not in db:
            missing_keys.append(k)

    if missing_keys:
        nfo += f'missing keys: {missing_keys}\n'

    new_keys = []
    new_values = []
    for k in db:
        if k not in da:
            new_keys.append(k)
        else:
            if da[k] != db[k]:
                new_values.append(k)

    if new_keys:
        nfo += f'new keys: {new_keys}\n'
        for k in new_keys:
            nfo += f' > {k}: {db[k]}\n'

    if new_values:
        nfo += f'new values: {new_values}\n'
        for k in new_values:
            nfo += f' > {k}: {db[k]}\n'

    if nfo: nfo = nfo[:-1]
    return nfo


class Para:
    """ Parameters Access
    its parameters (point) may be accessed in dict-like style []
    but Para IS NOT a dict
    protected fields may also be accessed with [], but rather should not """

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __contains__(self, key):
        return key in vars(self)

    # ********************************************************************************************************** getters

    def get_all_params(self, protected=False) -> List[str]:
        """ returns list of all self parameters """
        fields = [pm for pm in vars(self) if not pm.startswith('_Para__')]
        if not protected:
            fields = [pm for pm in fields if not pm.startswith('_')]
        return fields

    def get_managed_params(self) -> List[str]:
        """ prepares list of parameters managed by Para (all but private and protected) """
        return self.get_all_params(protected=False)

    def get_point(self) -> POINT:
        """ returns POINT of parameters managed by Para """
        return {k: self[k] for k in self.get_managed_params()}

    # *************************************************************************************************** update & check

    def update(self, d:dict) -> None:
        """ update in dict-like style """
        for key in d:
            self.__setitem__(key, d[key])

    def check_params_sim(
            self,
            params: Optional[Union[Dict,List]]= None,
            lev_dist_diff: int=                 1,
    ) -> Optional[Set[Tuple[str,str]]]:
        """ checks self.params + params for similarity against given
        returns True if already got similar """

        if params is None:
            params = []
        paramsL = params if type(params) is list else list(params.keys())

        self_paramsL = list(self.get_point().keys())
        diff_paramsL = [par for par in paramsL if par not in self_paramsL]

        all_params = sorted(list(set([p.lower() for p in self_paramsL + diff_paramsL])))

        #found_any = False
        found = set()
        for pa in all_params:
            for pb in all_params:
                if pa != pb:
                    levD = lev_dist(pa,pb)
                    if levD <= lev_dist_diff:
                        found.add(tuple(sorted([pa, pb])))

        return found or None


    def __str__(self):
        s = f'{self.__class__.__name__} (Para):\n'
        pms = self.get_point()
        if pms:
            for k in sorted(pms.keys()):
                s += f' > {k:30s}: {pms[k]}\n'
        else:
            s += f'---empty---\n'
        return s[:-1]


class ParaGX(Para):
    """ ParaGX - Para with GX (genetic crossing algorithm)
    implements GX on parents POINTs

    GX algorithm builds child POINT from:
        - parameters of parents POINTs
        - parameters sampled from merged both parents PSDD
    it is also possible to preform GX of SINGLE parent
    """

    def __init__(
            self,
            name: Optional[str]=    None,
            family: Optional[str]=  '__gx-fam__', # family of GXable, here default, for GX functionality family needs to be given
            psdd: Optional[PSDD]=   None, # PSDD of GXable
            **kwargs):

        if not name:
            name = self.__class__.__name__
        self.name = name

        self.family = family
        # INFO: all keys of self.psdd should be present in self
        #  - it is not checked now (while init) since self may be updated even after init, IT IS checked while GX
        self.psdd = psdd or {}

        # sample kwargs from psdd
        if not kwargs:

            if not psdd:
                raise PMSException('when ParaGX \'kwargs\' not given, \'psdd\' must be given')

            paspa = PaSpa(psdd=psdd, loglevel=30)
            kwargs = paspa.sample_point_GX()

        self.update(kwargs)

    @staticmethod
    def families_compatible(
            parentA: "ParaGX",
            parentB: Optional["ParaGX"]= None
    ) -> bool:
        """ checks for compatibility of families
        returns True when families match and are not None """

        pointA = parentA.get_point()
        fmA = pointA.get('family', None)
        pointB = parentB.get_point() if parentB else None

        if not pointB:
            return fmA is not None

        else:
            if fmA is None:
                return False

            fmB = pointB.get('family', None)
            return fmA == fmB

    @staticmethod
    def gx_point(
            parentA: "ParaGX",
            parentB: Optional["ParaGX"]=    None,
            name_child: Optional[str]=      None,
            prob_mix=                       0.5,
            prob_noise=                     0.3,
            noise_scale=                    0.1,
            prob_axis=                      0.1,
            prob_diff_axis=                 0.3
    ) -> POINT:
        """ GX on ParaGX POINTs """

        if not ParaGX.families_compatible(parentA, parentB):
            raise PMSException('ERR: parents families not compatible')

        pointA = parentA.get_point()
        pointB = parentB.get_point() if parentB else None

        psddA = pointA.get('psdd', {})
        psddB = pointB.get('psdd', {}) if pointB is not None else None
        psdd_merged = PaSpa.merge_psdd(
                psdd_a= psddA,
                psdd_b= psddB) if psddB is not None else psddA

        paspa_merged = PaSpa(psdd=psdd_merged, loglevel=30)

        paspa_axes_not_in_parent = [a for a in paspa_merged.axes if a not in pointA]
        if paspa_axes_not_in_parent:
            raise PMSException(f'ERR: PaSpa axes not in parentA: {paspa_axes_not_in_parent}')

        # sub-points limited to axes of psdd_merged (PaSpa.sample_point_GX() does not accept other axes..)
        pointA_psdd = {k: pointA[k] for k in psdd_merged}
        pointB_psdd = {k: pointB[k] for k in psdd_merged} if pointB is not None else None

        point_gx = paspa_merged.sample_point_GX(
            pointA=         pointA_psdd,
            pointB=         pointB_psdd,
            prob_mix=       prob_mix,
            prob_noise=     prob_noise,
            noise_scale=    noise_scale,
            prob_axis=      prob_axis,
            prob_diff_axis= prob_diff_axis)

        # point_child is based on parentA with updated values of point_gx
        point_child: POINT = deepcopy(pointA)
        point_child.update(point_gx)
        point_child['psdd'] = psdd_merged # update psdd to psdd_merged

        name_merged = f'{parentA.name}+{parentB.name}_(gxp)' if parentB else f'{parentA.name}_(sgxp)'
        point_child['name'] = name_child or name_merged

        return point_child

    @property
    def gxable_point(self) -> POINT:
        """ gxable_point is a self POINT limited to axes of self.psdd """
        return {k: self[k] for k in self.psdd}


    def __str__(self):
        s = f'{self.__class__.__name__} (ParaGX) name: {self.name}, family: {self.family}\n'
        pms = self.get_point()
        for k in sorted(pms.keys()):
            p = f'param: {k}'
            v = f'value: {pms[k]}'
            r = f'ranges:{self.psdd[k]}' if k in self.psdd else ''
            s += f'{p:30s} {v:30s} {r}\n'
        return s[:-1]


from matplotlib import pyplot as plt
import numpy as np
import os
import pandas as pd
import plotly.express as px
import scipy
from typing import List, Optional, Union

from pypaq.lipytools.files import prep_folder
from pypaq.lipytools.stats import stats_pd, msmx



def histogram(
        val_list: Union[List,np.ndarray],
        name=                   'values',
        rem_nstd: float=        0.0,    # removes values out of N*stddev
        msmx_stats=             True,   # prints minimal stats
        pandas_stats=           False,  # prints pandas extended stats
        density=                True,
        bins: Optional[int]=    None,   # automatic for None
        save_FD :str=           None):

    if type(val_list) is list and not val_list or type(val_list) is np.ndarray and not len(val_list):
        print(f'cannot prepare histogram for empty val_list!')
        return
    if msmx_stats: print(f' > "{name}": {msmx(val_list)["string"]}')
    if pandas_stats: print(f' > stats with pandas for "{name}":\n{stats_pd(val_list)}')

    if rem_nstd:
        stats = msmx(val_list)
        std = stats['std']
        mean = stats['mean']
        val_list = [val for val in val_list if mean - rem_nstd * std < val < mean + rem_nstd * std]

        if pandas_stats:
            print(f'\n > after removing {rem_nstd} stddev:')
            print(stats_pd(val_list))

    if not bins:
        bins = len(set(val_list))
        if bins>50: bins = 50

    plt.clf()
    n, x, _ = plt.hist(val_list, label=name, density=density, bins=bins, alpha=0.5)

    # try build density, for some val_list it is not possible
    try:
        density = scipy.stats.gaussian_kde(val_list)
        plt.plot(x, density(x))
    except: pass

    plt.legend(loc='upper right')
    plt.grid(True)
    if save_FD:
        if not os.path.isdir(save_FD): os.makedirs(save_FD, exist_ok=True)
        plt.savefig(f'{save_FD}/{name}.png')
    else:
        plt.show()


def two_dim(
        y: Union[List,np.ndarray],                  # two(yx) or one(y) dim list or np.array
        x: Optional[Union[List,np.ndarray]]=    None,
        name=                                   'values',
        save_FD: str =                          None,
        xlogscale=                              False,
        ylogscale=                              False,
        legend_loc=                             'upper left'):

    if type(y) is list: y = np.asarray(y)
    if x is None:
        if len(y.shape) < 2: x = np.arange(len(y))
        else:
            x = y[:, 1]
            y = y[:, 0]

    plt.clf()
    plt.plot(x, y, label=name)
    if xlogscale: plt.xscale('log')
    if ylogscale: plt.yscale('log')
    plt.legend(loc=legend_loc)
    plt.grid(True)
    if save_FD:
        prep_folder(save_FD)
        plt.savefig(f'{save_FD}/{name}.png')
    else:
        plt.show()


def two_dim_multi(
        ys: List[Union[List,np.ndarray]],
        names: Optional[List[str]]= None,
        name: Optional[str]=        None,
        save_FD: str=               None,
        xlogscale=                  False,
        ylogscale=                  False,
        legend_loc=                 'upper left'):

    if names is None:
        names = [f'values_{ix}' for ix in range(len(ys))]

    x = np.arange(len(ys[0]))

    plt.clf()
    for y,name in zip(ys,names):
        plt.plot(x, y, label=name)

    plt.legend(loc=legend_loc)
    plt.grid(True)
    if xlogscale: plt.xscale('log')
    if ylogscale: plt.yscale('log')

    if save_FD:
        prep_folder(save_FD)
        if name is None:
            name = "_".join(names)
        plt.savefig(f'{save_FD}/{name}.png')
    else:
        plt.show()


def three_dim(
    xyz: Union[list, np.ndarray], # sequence of (x,y,val) or (x,y,z,val)
    name=               'values',
    x_name=             'x',
    y_name=             'y',
    z_name=             'z',
    val_name=           'val',
    opacity=            0.7,
    width=              700,
    height=             700,
    save_FD: str =      None):

    # expand to 3 axes + val (3rd axis data)
    if len(xyz[0])<4:
        new_xyz = []
        for e in xyz: new_xyz.append(list(e) + [e[-1]])
        xyz = new_xyz

    df = pd.DataFrame(
        data=       xyz,
        columns=    [x_name,y_name,z_name,val_name])

    std = df[val_name].std()
    mean = df[val_name].mean()
    off = 2*std
    cr_min = mean - off
    cr_max = mean + off

    fig = px.scatter_3d(
        data_frame=     df,
        title=          name,
        x=              x_name,
        y=              y_name,
        z=              z_name,
        color=          val_name,
        range_color=    [cr_min,cr_max],
        opacity=        opacity,
        width=          width,
        height=         height)

    if save_FD:
        file = f'{save_FD}/{name}_3Dplot.html'
        fig.write_html(file, auto_open=False if os.path.isfile(file) else True)
    else: fig.show()


import csv
import json
import os
import pickle
import shutil
import yaml
from pathlib import Path
import sys
from typing import Union, Dict, List, Optional

from pypaq.exception import PyPaqException


def r_pickle(file_path, obj_type=None, raise_exception=False):
    """ reads pickle
    if obj_type is given checks for compatibility with given type """
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None

    # obj = pickle.load(open(file_path, 'rb')) << replaced by:
    with open(file_path, 'rb') as file: obj = pickle.load(file)

    if obj_type:
        if not type(obj) is obj_type:
            raise PyPaqException(f'ERROR: obj from file is not {str(obj_type)} type')
    return obj


def w_pickle(obj, file_path):
    with open(file_path, 'wb') as file:
        pickle.dump(obj, file)


def r_json(file_path, raise_exception=False) -> Optional[Union[Dict,List]]:
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path, 'r', encoding='utf-8') as file:
        return json.load(file)


def w_json(data:Union[Dict,List], file_path):
    with open(file_path, 'w', encoding='utf-8') as file:
        json.dump(data, file, indent=4, ensure_ascii=False)


def r_jsonl(file_path, raise_exception=False) -> Optional[List]:
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path, 'r', encoding='utf-8') as file:
        return [json.loads(line) for line in file]


def w_jsonl(data:List, file_path):
    with open(file_path, 'w', encoding='utf-8') as file:
        for d in data:
            json.dump(d, file, ensure_ascii=False)
            file.write('\n')


def r_csv(file_path, raise_exception=False) -> Optional[List]:
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    csv.field_size_limit(sys.maxsize)
    with open(file_path, newline='') as f:
        reader = csv.reader(f)
        return [row for row in reader][1:]


def r_yaml(file_path, raise_exception=False):
    if not os.path.isfile(file_path):
        if raise_exception: raise FileNotFoundError(f'file {file_path} not exists!')
        return None
    with open(file_path) as file:
        return yaml.load(file, yaml.Loader)


def get_dir(path: Union[str, Path]):
    path = str(path)
    path_split = path.split('/')
    if path_split[-1].find('.') != -1: path_split = path_split[:-1]
    return '/'.join(path_split)


def prep_folder(path:Union[str,Path], flush_non_empty=False):
    """ prepares folder
    path may be a file path -> folder path is extracted """
    folder_path = get_dir(path)
    if flush_non_empty and os.path.isdir(folder_path): shutil.rmtree(folder_path)
    os.makedirs(folder_path, exist_ok=True)


def list_dir(path: Union[str, Path]) -> Dict[str,List]:
    ls = os.listdir(path)
    lsD = {'files':[], 'dirs':[]}
    for e in ls:
        lsD['files' if os.path.isfile(f'{path}/{e}') else 'dirs'].append(e)
    return lsD


def get_requirements(fileFP:str='requirements.txt') -> List[str]:
    with open(fileFP) as file:
        return [l.strip() for l in file]




import math
from typing import Sized, List, Tuple, Optional, Dict, Union

from pypaq.lipytools.pylogger import get_pylogger
from pypaq.lipytools.stats import mam
from pypaq.lipytools.plots import three_dim
from pypaq.lipytools.printout import float_to_str
from pypaq.pms.base import POINT, point_str
from pypaq.pms.paspa import PaSpa


class VPoint:
    """ Valued Point
    with unique id & float value """

    def __init__(
            self,
            point: POINT,
            name: Optional[str]=    None,
            id: Optional[int]=      None,
            value: Optional[float]= None,
    ):
        self.point = point
        self.name = name
        self.id = id
        self.value = value

    def __str__(self):
        nfo = self.name if self.name else ""
        if not nfo:
            nfo = f'#{self.id:4}' if self.id is not None else '#_'
        val_nfo = f' [val: {self.value:.8f}] ' if self.value is not None else ''
        return f'{nfo}{val_nfo}{point_str(self.point)}'


def points_nice_table(
        vpoints: List[VPoint],
        do_name: bool=  True,
        do_val: bool=   True,
) -> List[str]:
    """ prepares list of str with nice formatted VPoints params and values
    points may come from different spaces, their given order is kept in returned list """

    keys = []
    for vp in vpoints:
        keys += list(vp.point.keys())
    keys = sorted(list(set(keys)))

    # key width
    kw = {k: len(k) for k in keys}
    kw['_name'] = 4
    kw['_val'] = 3

    print_vals = []
    for vp in vpoints:

        pv = {}
        print_vals.append(pv)
        for k in keys:

            val = vp.point[k] if k in vp.point else ''

            val_str = ''
            if type(val) is float:
                val_str = float_to_str(val, fill=False)
            if type(val) is bool:
                val_str = f'{str(val):5}'
            if not val_str:
                val_str = str(val)

            pv[k] = val_str

            lv = len(val_str)
            if lv > kw[k]:
                kw[k] = lv

        if vp.name:
            ln = len(vp.name) if vp.name else 0
            if ln > kw['_name']:
                kw['_name'] = ln

        if vp.value:
            lv = len(float_to_str(vp.value, fill=False))
            if lv > kw['_val']:
                kw['_val'] = lv

    s = ''
    if do_name: s += 'name' + ' ' * (kw['_name']-3)
    if do_val:  s += 'val'  + ' ' * (kw['_val']-2)
    s += ' '.join([f'{k:{kw[k]}}' for k in keys])
    table = [s]
    for vp,pv in zip(vpoints,print_vals):
        name = f'{vp.name:{kw["_name"]}}' if vp.name else ' ' * kw["_name"]
        val = f'{float_to_str(vp.value, fill=False):{kw["_val"]}}' if vp.value is not None else ' ' * kw["_val"]
        s = ''
        if do_name: s += f'{name} '
        if do_val:  s += f'{val} '
        s += ' '.join([f'{pv[k]:{kw[k]}}' for k in keys])
        table.append(s)

    return table


class PointsCloud(Sized):
    """ Points Cloud - cloud of Valued Points """

    def __init__(
            self,
            paspa: PaSpa,   # space of this PointsCloud
            logger=     None,
            loglevel=   20):

        if not logger:
            logger = get_pylogger(name=self.__class__.__name__, level=loglevel)
        self.logger = logger
        self.logger.info('*** PointsCloud *** initializing..')

        self.paspa = paspa

        # values below are updated with each call to update_cloud()
        self._vpointsD: Dict[int, VPoint] = {}          # {id: VPoint}
        self._nearest: Dict[int, Tuple[int,float]] = {} # {id: (id,dist)}
        self.min_nearest = math.sqrt(self.paspa.dim)
        self.avg_nearest = math.sqrt(self.paspa.dim)
        self.max_nearest = math.sqrt(self.paspa.dim)

        self.prec = 8 # print precision, will be updated while adding new vpoints


    def distance(self, vpa:VPoint, vpb:VPoint) -> float:
        """ returns distance between two VPoints """
        return self.paspa.distance(vpa.point, vpb.point)


    def update_cloud(self, vpoints:Union[VPoint,List[VPoint]]):
        """ updates Cloud (self) with given VPoint / list
        (adds new to _vpoints & updates _nearest) """

        if vpoints:

            if type(vpoints) is not list:
                vpoints = [vpoints]

            for vpoint in vpoints:

                # add to _vpoints
                vp_id = len(self)
                vpoint.id = vp_id
                self._vpointsD[vp_id] = vpoint

                # update _nearest
                his_nearest = None
                his_nearest_dist = None
                for k in self._vpointsD:
                    if k != vp_id:
                        dist = self.distance(vpoint, self._vpointsD[k])
                        if his_nearest is None or dist < his_nearest_dist:
                            his_nearest = k
                            his_nearest_dist = dist
                        if k not in self._nearest or dist < self._nearest[k][1]:
                            self._nearest[k] = vp_id, dist
                if his_nearest is not None:
                    self._nearest[vp_id] = his_nearest, his_nearest_dist

                if vpoint.value is not None and vpoint.value > 0.01: self.prec = 4

            self.min_nearest, self.avg_nearest, self.max_nearest = mam([v[1] for v in self._nearest.values()])


    def plot(
            self,
            name: str=                  'PointsCloud',
            axes: Optional[List[str]]=  None,   # list with axes names, 2-3, like ['drop_a','drop_b','loss']
            folder: Optional[str]=      None,
    ):
        """ prepares 3D plot of the Cloud VPoints """

        columns = sorted(list(self._vpointsD[0].point.keys()))[:3] if not axes else [] + axes

        if len(columns) < 2:
            self.logger.warning('Cannot prepare 3D plot for less than two axes')

        else:

            valLL = [[sp.point[key] for key in columns] for sp in self._vpointsD.values()]

            # eventually add score
            if len(columns) < 4:
                columns += ['value']
                valLL = [vl + [sp.value] for vl,sp in zip(valLL, self._vpointsD.values())]
            three_dim(
                xyz=        valLL,
                name=       name,
                x_name=     columns[0],
                y_name=     columns[1],
                z_name=     columns[2],
                val_name=   'val',
                save_FD=    folder)

    @property
    def vpoints(self) -> List[VPoint]:
        return list(self._vpointsD.values())

    def __len__(self):
        """ number of VPoints in the Cloud """
        return len(self._vpointsD)

    def __str__(self):
        vpoints = self.vpoints
        if None not in [vp.value for vp in vpoints]:
            vpoints.sort(key=lambda x: x.value, reverse=True)
        return '\n'.join(points_nice_table(vpoints))



import numpy as np
from typing import Union, Sequence


NUM = Union[int, float, np.ndarray]     # just one number, in case of np.ndarray it has shape ()
NPL = Union[Sequence[NUM], np.ndarray]  # array of numbers


from setuptools import setup, find_packages

from pypaq.lipytools.files import get_requirements


setup(
    name=               'pypaq',
    version=            'v1.6.2',
    url=                'https://github.com/piteren/pypaq.git',
    author=             'Piotr Niewinski',
    author_email=       'pioniewinski@gmail.com',
    description=        'little Python tools',
    packages=           find_packages(),
    install_requires=   get_requirements(),
    license=            'MIT')


from copy import deepcopy
import math
import random
from typing import Optional, Dict, Tuple, Union

from pypaq.pms.base import P_VAL, POINT, PSDD, PMSException
from pypaq.lipytools.pylogger import get_pylogger, get_child

NO_REF = '<<< NO-REF >>>' # axis value when no reference value given (since 'None' cannot be used)



class PaSpa:
    """ PaSpa - Parameters Space

    PaSpa is build from PSDD (Params Space Definition Dictionary),
    PaSpa has two properties that may describe its complexity:
    - dim  - dimensionality (= number of axes)
    - rdim - reduced dimensionality (rdim<=dim - since some axes are simpler(tuples, lists of ints))

    Each axis of PaSpa has:
    - type (list, tuple) of (float, int, diff)
    - range (width)

    PaSpa is a metric space, supports L2 distance calculations, normalized to 1 (the longest space diagonal) """

    def __init__(
            self,
            psdd:PSDD,
            seed:Optional[int]= None,
            logger=             None,
            loglevel=           30,
    ):

        if not logger:
            logger = get_pylogger(name=self.__class__.__name__, level=loglevel)
        self.logger = logger

        self._psdd = psdd
        self.axes = sorted(list(self._psdd.keys()))

        if seed is not None:
            random.seed(seed)

        self._axT, self._axW = self.__axes_type_width()

        # width of str(value) for axes, used for nice str formatting
        self._str_width = {}
        for axis in self.axes:
            width = max([len(str(e)) for e in self._psdd[axis]])
            if 'list_float' in self._axT[axis]: width = 7
            self._str_width[axis] = width

        self.logger.info(f'*** PaSpa *** initialized, dim: {self.dim},rdim: {self.rdim:.1f}')

    ### *************************************************************************** axes / value_on_axis related methods

    def __axes_type_width(self) -> Tuple[Dict,Dict]:
        """ prepares axes type and width + some safety checks
        possible axes types:
        - list_float
        - list_int
        - tuple_float
        - tuple_int
        - tuple_diff """

        axT: Dict[str, str] = {}   # axis type
        axW: Dict[str, float] = {} # axis width, range (min <-> max), for diff_tuple - number of elements

        for axis in self.axes:

            if not type(self._psdd[axis]) in (list, tuple): self._psdd[axis] = (self._psdd[axis],) # replace P_VAL by (P_VAL,)

            pdef = self._psdd[axis]
            tp = 'tuple' if type(pdef) is tuple else 'list'
            if tp == 'list' and len(set(pdef))!= 2:
                raise PMSException(f'parameter definition with list should have two different elements: >{pdef}<!')

            tpn = 'int' # default
            are_flt = False
            are_dif = False
            for el in pdef:
                if type(el) is float: are_flt = True
                if type(el) is not int and type(el) is not float: are_dif = True
            if are_flt: tpn = 'float'  # downgrade
            if are_dif: tpn = 'diff'   # downgrade

            if tp=='list' and tpn=='diff':
                raise PMSException(f'ERR: axis {axis} defined with list may contain only floats or ints')

            # sort numeric
            if tpn != 'diff':
                pdef = sorted(list(pdef))
                self._psdd[axis] = pdef if tp == 'list' else tuple(pdef) # rollback type of sorted

            axT[axis] = f'{tp}_{tpn}'
            axW[axis] = pdef[-1] - pdef[0] if tpn != 'diff' else len(pdef) - 1

        return axT, axW

    def __closest(
            self,
            ref_val: Union[float, int],      # for diff_tuple it is index of value in self.__psdd[axis]
            axis: str,
    ) -> Union[float, int]:
        """ selects value from axis closest to given val """

        val = ref_val  # for list_float type

        axT = self._axT[axis]

        if axT == 'list_int':
            val = int(round(val))

        if 'tuple' in axT:

            # num tuple
            if 'diff' not in axT:
                axis_dst = [(e, abs(e - ref_val)) for e in self._psdd[axis]]
                axis_dst.sort(key=lambda x: x[1])
                val = axis_dst[0][0]

            # diff_tuple - value of the closest index
            else: val = self._psdd[axis][int(round(val))]

        if not self.__value_in_axis(value=val, axis=axis):
            raise PMSException(f'something went wrong since val {val} is not in {axis}')

        return val

    @staticmethod
    def __apply_noise(
            ref_val: Union[float, int], # reference value (in range og L & R)
            noise_scale: float,         # <0.0;1.0> distance from ref_val as a factor of range where new val will be sampled
            rngL: Union[float, int],    # range Left value
            rngR: Union[float, int],    # range Right value
    ) -> float:
        """ applies noise to given ref_val
        (samples new value from sub-range defined by noise scale) """

        dist = noise_scale * (rngR-rngL)

        # left side of ref_val
        if random.random() < 0.5:
            rng = ref_val - dist        # new left range
            if rng < rngL: rng = rngL   # limit to given range

        # right size
        else:
            rng = ref_val + dist        # new right range
            if rng > rngR: rng = rngR   # limit to given range

        return random.uniform(ref_val, rng)

    def __random_value_noref(self, axis:str) -> P_VAL:
        """ samples random value from whole axis """

        axT = self._axT[axis]
        psdd = self._psdd[axis]

        if 'list' in axT:
            a = psdd[0]
            b = psdd[1]
            val = random.uniform(a,b)
            if 'int' in axT: val = int(round(val))
        else: val = random.choice(psdd)

        return val

    def __random_value(
            self,
            axis: str,
            ref_val: P_VAL,         # reference value on axis, possible NO_REF
            prob_noise: float,      # probability of shifting value with noise (for list or tuple_with_num)
            noise_scale: float,     # max noise scale (axis width, for list or tuple_with_num)
            prob_axis: float,       # probability of value replacement by one sampled from whole axis (for list or num_tuple)
            prob_diff_axis:float    # probability of value replacement by one sampled from whole axis (for diff_tuple)
    ) -> P_VAL:
        """ gets random value for axis
        (algorithm ensures equal probability for both sides of ref_val) """

        if ref_val == NO_REF:
            return self.__random_value_noref(axis)

        axT = self._axT[axis]
        axis_psd = self._psdd[axis]

        # list or num_tuple
        if 'diff' not in axT:

            # whole axis
            if random.random() < prob_axis:
                return self.__random_value_noref(axis)

            val = ref_val
            # add noise
            if noise_scale and random.random() < prob_noise:
                val = PaSpa.__apply_noise(
                    ref_val=        ref_val,
                    noise_scale=    noise_scale,
                    rngL=           axis_psd[0],
                    rngR=           axis_psd[-1])
                val = self.__closest(val, axis)

        # tuple_diff
        else:

            # whole axis
            if random.random() < prob_diff_axis:
                return self.__random_value_noref(axis)

            val = ref_val
            # add noise (on indexes)
            if noise_scale and random.random() < prob_noise:
                val_ix = axis_psd.index(val)
                val_ix = PaSpa.__apply_noise(
                        ref_val=        val_ix,
                        noise_scale=    noise_scale,
                        rngL=           0,
                        rngR=           len(axis_psd)-1)
                val = self.__closest(val_ix, axis)

        return val


    def __value_in_axis(self, value:P_VAL, axis:str) -> bool:
        """ checks if given value belongs to an axis of this space """
        if axis not in self._axT:                                      return False # axis not in a space
        if 'list' in self._axT[axis]:
            if type(value) is float and 'int' in self._axT[axis]:      return False # type mismatch
            if self._psdd[axis][0] <= value <= self._psdd[axis][1]:   return True
            else:                                                       return False # value not in a range
        elif value not in self._psdd[axis]:                            return False # value not in a tuple
        return True

    ### *************************************************************************** "point in the space" related methods

    def sample_point(
            self,
            ref_point: Optional[POINT]= None,   # reference point
            prob_noise=                 0.3,    # probability of shifting value with noise (for list or num_tuple)
            noise_scale=                0.1,    # max noise scale (axis width)
            prob_axis=                  0.1,    # probability of value replacement by one sampled from whole axis (for list or num_tuple)
            prob_diff_axis=             0.3     # probability of value replacement by one sampled from whole axis (for diff_tuple)
    ) -> POINT:
        """ samples (random) point from whole space or from the surroundings of ref_point """
        if ref_point is None:
            ref_point = {axis: NO_REF for axis in self.axes}
        return {axis: self.__random_value(
            axis=           axis,
            ref_val=        ref_point[axis],
            prob_noise=     prob_noise,
            noise_scale=    noise_scale,
            prob_axis=      prob_axis,
            prob_diff_axis= prob_diff_axis) for axis in self.axes}

    def sample_point_GX(
            self,
            pointA: Optional[POINT]=    None,   # parent A, when not given > samples from the whole space
            pointB: Optional[POINT]=    None,   # parent B, when not given > samples from the surroundings of pointA
            prob_mix=                   0.5,    # probability of mixing two values (for list or num_tuple - but only when both parents are given)
            prob_noise=                 0.3,    # probability of shifting value with noise (for list or num_tuple)
            noise_scale=                0.2,    # max noise scale (axis width)
            prob_axis=                  0.1,    # probability of value replacement by one sampled from whole axis (for list or num_tuple)
            prob_diff_axis=             0.3     # probability of value replacement by one sampled from whole axis (for diff_tuple)
    ) -> POINT:
        """ samples GX point from given None, one or two """

        ref_point = None # when pointA nor pointB are given

        if pointA:
            if not self.is_from_space(pointA):
                raise PMSException(f'ERR: pointA: {pointA} not from space: {self._psdd}')

            ref_point = pointA  # select main parent

            # build ref_point with mix or select
            if pointB:
                if not self.is_from_space(pointB):
                    raise PMSException(f'pointB: {pointB} not from space: {self._psdd}')

                ref_point = {}
                ratio = 0.5 + random.random() / 2  # mix/select ratio of parentA, rest from parentB (at least half from parentA)

                for axis in self.axes:

                    # mix
                    if random.random() < prob_mix:

                        # num
                        if 'diff' not in self._axT[axis]:
                            val = pointA[axis] * ratio + pointB[axis] * (1 - ratio)
                            val = self.__closest(val, axis)
                        # diff_tuple - mix on indexes
                        else:
                            pm_ix = self._psdd[axis].index(pointA[axis])
                            ps_ix = self._psdd[axis].index(pointB[axis])
                            val_ix = pm_ix * ratio + ps_ix * (1-ratio)
                            val = self.__closest(val_ix, axis)
                    
                    # select
                    else: val = pointA[axis] if random.random() < ratio else pointB[axis]

                    ref_point[axis] = val

        # sample from the surroundings of ref_point
        return self.sample_point(
            ref_point=      ref_point,
            prob_noise=     prob_noise,
            noise_scale=    noise_scale,
            prob_axis=      prob_axis,
            prob_diff_axis= prob_diff_axis)

    def sample_corners(self) -> Tuple[POINT, POINT]:
        """ samples 2 corner points with max distance (1 in normalized space) """
        pa = {}
        pb = {}
        left = [0 if random.random()>0.5 else 1 for _ in range(self.dim)] # left/right
        for aIX in range(self.dim):
            ax = self.axes[aIX]
            vl = self._psdd[ax][0]
            vr = self._psdd[ax][-1]
            pa[ax] = vl
            pb[ax] = vr
            if left[aIX]:
                pa[ax] = vr
                pb[ax] = vl
        return pa, pb

    def is_from_space(self, point:POINT) -> bool:
        """ checks if given point comes from this space """
        if set(point.keys()) != set(self.axes): return False
        for axis in point:
            if not self.__value_in_axis(value=point[axis], axis=axis):
                return False
        return True

    def distance(self, pa:POINT, pb:POINT) -> float:
        """ L2 distance between two points of this space (normalized) """
        dist_pow_sum = 0
        for axis in pa:
            if self._axW[axis] > 0:
                dist = self._psdd[axis].index(pa[axis]) - self._psdd[axis].index(pb[axis]) \
                    if 'diff' in self._axT[axis] else \
                    pa[axis] - pb[axis]
                dist_pow_sum += (dist / self._axW[axis]) ** 2
        return  math.sqrt(dist_pow_sum) / math.sqrt(self.dim)

    def point_normalized(self, p:POINT) -> POINT:
        """ prepares normalized point of p
        value of each axis is represented by a float <0.0;1.0> """

        pn: Dict[str,float] = {}

        for axis in p:

            if self._axT[axis] == 'list_float':
                pn[axis] = (p[axis] - self._psdd[axis][0]) / self._axW[axis] # position in float range

            else:

                # |_._|_._|_._| <- each point is placed in the middle of _ _ (sub width)
                if self._axT[axis] == 'list_int':
                    num_elements = self._axW[axis] + 1
                    v_ix = p[axis] - self._psdd[axis][0]
                else:
                    num_elements = len(self._psdd[axis])
                    v_ix = self._psdd[axis].index(p[axis])

                sw = 1 / num_elements # sub width
                pn[axis] = sw/2 + v_ix*sw # half of sw + 1/2 sw

        return pn

    ### *********************************************************************************** space methods and properties

    # merges two PSDD
    @staticmethod
    def merge_psdd(
            psdd_a: PSDD,
            psdd_b: PSDD) -> PSDD:
        return (PaSpa(psdd_a, loglevel=30) + PaSpa(psdd_b, loglevel=30)).psdd

    @property
    def psdd(self) -> PSDD:
        """ returns a copy of self._psdd """
        return deepcopy(self._psdd)

    @property
    def dim(self) -> int:
        return len(self._psdd)

    @property
    def rdim(self) -> float:
        """ calculates reduced dimensionality of PaSpa
        rdim = log10(∏ sq if sq<10 else 10) for all axes
            sq = 10 for list of floats (axis)
            sq = sqrt(len(axis_elements)) for tuple or list of int """
        axd = []
        for axis in self.axes:
            axt = self._axT[axis]
            if 'list' in axt: sq = 10 if 'float' in axt else math.sqrt(self._axW[axis])
            else:             sq = math.sqrt(len(self._psdd[axis]))
            axd.append(sq if sq<10 else 10)
        mul = 1
        for e in axd: mul *= e
        return math.log10(mul)

    @property
    def n_points(self) -> Optional[int]:
        """ returns number of points in PaSpa
        if PaSpa has infinite number of points -> returns None """
        n = 1
        for axis in self._axT:
            if self._axT[axis] == 'list_float':
                return None
            else:
                if self._axT[axis] == 'list_int':
                    n *= self._axW[axis] + 1
                else:
                    n *= len(self._psdd[axis])
        return n

    def __eq__(self, other):
        """ -> same axes, same definitions, same L """
        if self.axes != other.axes: return False
        for k in self._psdd.keys():
            if self._psdd[k] != other._psdd[k]:
                return False
        return True

    def __add__(self, other:"PaSpa") -> "PaSpa":
        psdd_a = self.psdd
        psdd_b = other.psdd
        psdd_merged = {}
        psdd_merged.update(psdd_a)
        for ax in other.axes:
            if ax not in psdd_merged:
                psdd_merged[ax] = psdd_b[ax]
            else:
                if not type(psdd_merged[ax]) == type(psdd_b[ax]):
                    raise PMSException(f'ERR: types of axes \'{ax}\' in both PaSpa differs')
                if 'tuple' in other._axT[ax]:
                    # add new elements to the end
                    if 'diff' in other._axT[ax]:
                        merged = list(psdd_merged[ax])
                        for e in psdd_b[ax]:
                            if e not in merged: merged.append(e)
                    else:
                        merged = list(set(list(psdd_merged[ax]) + list(psdd_b[ax])))
                        merged.sort()
                    psdd_merged[ax] = tuple(merged)
                else:
                    ranges = psdd_merged[ax] + psdd_b[ax]
                    psdd_merged[ax] = [min(ranges), max(ranges)]
        return PaSpa(psdd_merged, logger=get_child(self.logger))

    def __str__(self):
        info = f'*** PaSpa *** (dim: {self.dim}, rdim: {self.rdim:.1f}) parameters space:\n'
        max_ax_l = 0
        max_ps_l = 0
        for axis in self.axes:
            if len(axis)                > max_ax_l: max_ax_l = len(axis)
            if len(str(self._psdd[axis])) > max_ps_l: max_ps_l = len(str(self._psdd[axis]))
        if max_ax_l > 40: max_ax_l = 40
        if max_ps_l > 70: max_ps_l = 70

        for axis in self.axes:
            info += f'> {axis:{max_ax_l}s}  {str(self._psdd[axis]):{max_ps_l}s}  {self._axT[axis]:11s}  width: {self._axW[axis]}\n'

        if self.n_points:
            info += f'number of points in space: {self.n_points}\n'

        return info[:-1]


from pypaq.exception import PyPaqException
from pypaq.pytypes import NUM
from typing import Optional



class MovAvg:
    """ moving average
    updates self.value with factor (given while init) """

    def __init__(
            self,
            factor: NUM=        0.1,    # (0.0;1.0>
            first_avg: bool=    True,   # first 1/factor values will be averaged
    ):
        self.value: Optional[NUM] = None

        if not 0 < factor <= 1:
            raise PyPaqException('factor should: 0 < factor <= 1')

        self.factor = factor
        self.upd_ix = 0
        self.first_avg = first_avg
        self.firstL = []


    def upd(self, val:NUM):

        if self.first_avg and self.upd_ix < 1/self.factor:
            self.firstL.append(val)
            self.value = sum(self.firstL) / len(self.firstL)
        else:
            if self.value is None: self.value = val
            else: self.value = (1-self.factor)*self.value + self.factor*val

        self.upd_ix += 1

        return self.value


    def reset(self, val:Optional[NUM]=None):
        self.value = val
        self.upd_ix = 0
        self.firstL = []


    def __call__(self) -> NUM:
        if self.value is None:
            raise PyPaqException('MovAvg not updated yet, value unknown')
        return self.value